/*
 *  This file was rewrited to enable multifile capture (over 2G)
 *
 *  Copyright (C) 1999 - 2000 Mogi, Kazhiro <ee69011@isc.meiji.ac.jp>
 *
 * based on: 
 *
 *  Some utilities for writing and reading AVI files.
 *  These are not intended to serve for a full blown
 *  AVI handling software (this would be much too complex)
 *  The only intention is to write out MJPEG encoded
 *  AVIs with sound and to be able to read them back again.
 *
 *  Copyright (C) 1999 Rainer Johanni <Rainer@Johanni.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <fcntl.h>

#include "avilib.h"

/*******************************************************************
 *                                                                 *
 *    Utilities for writing an AVI File                            *
 *                                                                 *
 *******************************************************************/

#define MAKEFOURCC(a,b,c,d) ( (a) | ((b)<< 8) | ((c)<<16) | ((d)<<24) )

long tell(int fd)
{
   return (lseek(fd,0,SEEK_CUR));
}

out4cc (int fd, char *s)
{
  if( write(fd,s,4) != 4 ) { perror("write"); exit(1); }
}
outlong(int fd, long n)
{
  long s = n;

  if( write(fd,&s,4) != 4 ) { perror("write"); exit(1); }
}
outshrt(int fd, long n)
{
  short s = n;

  if( write(fd,&s,2) != 2 ) { perror("write"); exit(1); }
}

/* Handling of lists */

void AVI_start_list(AVI_file *af, char * name)
{

  af->list_depth++;
  if(af->list_depth > MAX_LIST_DEPTH)
  {
    fprintf(stderr," MAX_LIST_DEPTH exceeded\n");
    exit(1);
  }

  out4cc (af->fd, "LIST");
  outlong(af->fd, 0); /* Length of list in bytes, will be written
                        by AVI_finish_list */
  af->list_pos[af->list_depth-1] = tell(af->fd);
  out4cc (af->fd, name);
  strncpy(af->list_name[af->list_depth-1],name,4);

}

void AVI_finish_list(AVI_file *af, char * name)
{

  int pos;

  if(af->list_depth < 1)
  {
    fprintf(stderr,"AVI_finish_list called without a list started\n");
    exit(1);
  }

  if(strncmp(name,af->list_name[af->list_depth-1],4) )
  {
    fprintf(stderr,"AVI_finish_list does not match open list\n");
    exit(1);
  }

  pos = tell(af->fd);
  lseek(af->fd, af->list_pos[af->list_depth -1] -4, SEEK_SET);
  outlong(af->fd, pos - af->list_pos[af->list_depth -1]);
  lseek(af->fd,pos,SEEK_SET);

  af->list_depth--;
}
  
AVI_file AVI_open_output_file(char * filename)
{

   int i;
   char c[DATASTART-12];
   AVI_file af;

   af.fd = open(filename,O_RDWR|O_CREAT|O_TRUNC,0600);
   if (af.fd < 0) { perror("open"); exit(1); }

   /* Write out DATASTART - 12 bytes, the header will go here
      when we are finished with writing */

   for (i=0;i<DATASTART-12;i++) c[i] = 0;
   write(af.fd,c,DATASTART-12);

   return af;
}


/*
  Write the header of an AVI file.
  This routine must be called as the LAST routine !!!!!
  (It needs to know the file size);
*/

#ifndef VIDEO_MODE_PAL
/* from <linux/videodev.h> */
#define VIDEO_MODE_PAL          0
#define VIDEO_MODE_NTSC         1
#endif


void AVI_write_header(AVI_file *af, int width, int height, int norm,
				 int audio, int stereo, int size, int rate)
{

   long nbytes, pos, njunk, abytes, achans, sampsize;

   /* write index */

   lseek(af->fd,0,SEEK_END);
   out4cc(af->fd, "idx1");
   outlong(af->fd, af->n_idx*4);       /* # of bytes to follow */
   nbytes = write(af->fd,af->idx,af->n_idx*4);
   if (nbytes != af->n_idx*4) { perror("write"); exit(1); }

   /* get the file size */

   lseek(af->fd,0,SEEK_END);
   nbytes = tell(af->fd);
   lseek(af->fd,0,SEEK_SET); /* Rewind */

   if(!audio && af->num_of_audio_bytes)
      fprintf(stderr,"AVI: Audio samples present, but no audio specified???\n");

   /* The RIFF header */

   out4cc (af->fd, "RIFF");
   outlong(af->fd, nbytes - 8);    /* # of bytes to follow */
   out4cc (af->fd, "AVI ");

   /* Start the header list */

   AVI_start_list(af, "hdrl");

   /* The main AVI header */

   /* The Flags in AVI File header */

#define AVIF_HASINDEX           0x00000010      // Index at end of file
#define AVIF_MUSTUSEINDEX       0x00000020
#define AVIF_ISINTERLEAVED      0x00000100
#define AVIF_TRUSTCKTYPE        0x00000800      // Use CKType to find key frames
#define AVIF_WASCAPTUREFILE     0x00010000
#define AVIF_COPYRIGHTED        0x00020000

   out4cc (af->fd, "avih");
   outlong(af->fd, 56);            /* # of bytes to follow */
   if(norm==VIDEO_MODE_PAL)
      outlong(af->fd, 40000);      /* Microseconds per frame */
   else
      outlong(af->fd, 33367);      /* Microseconds per frame, NTSC, might need adjustment */
   outlong(af->fd, 10000000);      /* MaxBytesPerSec, I hope this will never be used */
   outlong(af->fd, 0);             /* PaddingGranularity (whatever that might be) */
   outlong(af->fd, 0x10010);       /* Flags: AVIF_HASINDEX | AVIF_WASCAPTUREFILE */
   outlong(af->fd, af->num_of_frames);    /* TotalFrames */
   outlong(af->fd, 0);             /* InitialFrames */
   if (af->num_of_audio_bytes)
      outlong(af->fd, 2);          /* Streams */
   else
      outlong(af->fd, 1);          /* Streams */
   outlong(af->fd, 0);             /* SuggestedBufferSize */
   outlong(af->fd, width);         /* Width */
   outlong(af->fd, height);        /* Height */
                                   /* MS calls the following 'reserved': */
   outlong(af->fd, 0);             /* TimeScale:  Unit used to measure time */
   outlong(af->fd, 0);             /* DataRate:   Data rate of playback     */
   outlong(af->fd, 0);             /* StartTime:  Starting time of AVI data */
   outlong(af->fd, 0);             /* DataLength: Size of AVI data chunk    */


   /* Start the video stream list ---------------------------------- */

   AVI_start_list(af, "strl");

   /* The video stream header */

   out4cc (af->fd, "strh");
   outlong(af->fd, 64);            /* # of bytes to follow */
   out4cc (af->fd, "vids");
   out4cc (af->fd, "MJPG");
   outlong(af->fd, 0);             /* Flags */
   outlong(af->fd, 0);             /* Reserved, MS says: wPriority, wLanguage */
   outlong(af->fd, 0);             /* InitialFrames */
   if(norm==VIDEO_MODE_PAL)
      outlong(af->fd, 40000);      /* Scale */
   else
      outlong(af->fd, 33367);      /* Scale */
   outlong(af->fd, 1000000);       /* Rate: Rate/Scale == samples/second */
   outlong(af->fd, 0);             /* Start */
   outlong(af->fd, af->num_of_frames);    /* Length */
   outlong(af->fd, 0);             /* SuggestedBufferSize */
   outlong(af->fd, -1);            /* Quality */
   outlong(af->fd, 0);             /* SampleSize */
   outlong(af->fd, 0);             /* Frame */
   outlong(af->fd, 0);             /* Frame */
   outlong(af->fd, 0);             /* Frame */
   outlong(af->fd, 0);             /* Frame */

   /* The video stream format */

   out4cc (af->fd, "strf");
   outlong(af->fd, 68);            /* # of bytes to follow */
   outlong(af->fd, 68);            /* Size */
   outlong(af->fd, width);         /* Width */
   outlong(af->fd, height);        /* Height */
   outshrt(af->fd, 1); outshrt(af->fd, 24);/* Planes, Count */
   out4cc (af->fd, "MJPG");        /* Compression */
   outlong(af->fd, width*height);  /* SizeImage ??? */
   outlong(af->fd, 0);             /* ??? */
   outlong(af->fd, 0);             /* ??? */
   outlong(af->fd, 0);             /* ??? */
   outlong(af->fd, 0);             /* ??? */
   outlong(af->fd, 44);            /* ??? */
   outlong(af->fd, 24);            /* ??? */
   outlong(af->fd, 0);             /* ??? */
   outlong(af->fd, 2);             /* ??? */
   outlong(af->fd, 8);             /* ??? */
   outlong(af->fd, 2);             /* ??? */
   outlong(af->fd, 1);             /* ??? */

   /* Finish stream list */

   AVI_finish_list(af, "strl");

   if (af->num_of_audio_bytes)
   {

   abytes = (size==16) ? 2 : 1;
   achans = stereo ? 2 : 1;
   sampsize = abytes*achans;

   /* Start the audio stream list ---------------------------------- */

   AVI_start_list(af, "strl");

   /* The audio stream header */

   out4cc (af->fd, "strh");
   outlong(af->fd, 64);            /* # of bytes to follow */
   out4cc (af->fd, "auds");
   out4cc (af->fd, "\0\0\0\0");
   outlong(af->fd, 0);             /* Flags */
   outlong(af->fd, 0);             /* Reserved, MS says: wPriority, wLanguage */
   outlong(af->fd, 0);             /* InitialFrames */
   outlong(af->fd, sampsize);      /* Scale */
   outlong(af->fd, sampsize*rate); /* Rate: Rate/Scale == samples/second */
   outlong(af->fd, 0);             /* Start */
   outlong(af->fd, af->num_of_audio_bytes/sampsize);   /* Length */
   outlong(af->fd, 0);             /* SuggestedBufferSize */
   outlong(af->fd, -1);            /* Quality */
   outlong(af->fd, sampsize);      /* SampleSize */
   outlong(af->fd, 0);             /* Frame */
   outlong(af->fd, 0);             /* Frame */
   outlong(af->fd, 0);             /* Frame */
   outlong(af->fd, 0);             /* Frame */

#define WAVE_FORMAT_UNKNOWN             (0x0000)
#define WAVE_FORMAT_PCM                 (0x0001)
#define WAVE_FORMAT_ADPCM               (0x0002)
#define WAVE_FORMAT_IBM_CVSD            (0x0005)
#define WAVE_FORMAT_ALAW                (0x0006)
#define WAVE_FORMAT_MULAW               (0x0007)
#define WAVE_FORMAT_OKI_ADPCM           (0x0010)
#define WAVE_FORMAT_DVI_ADPCM           (0x0011)
#define WAVE_FORMAT_DIGISTD             (0x0015)
#define WAVE_FORMAT_DIGIFIX             (0x0016)
#define WAVE_FORMAT_YAMAHA_ADPCM        (0x0020)
#define WAVE_FORMAT_DSP_TRUESPEECH      (0x0022)
#define WAVE_FORMAT_GSM610              (0x0031)
#define IBM_FORMAT_MULAW                (0x0101)
#define IBM_FORMAT_ALAW                 (0x0102)
#define IBM_FORMAT_ADPCM                (0x0103)

   /* The audio stream format */

   out4cc (af->fd, "strf");
   outlong(af->fd, 16);            /* # of bytes to follow */
   outshrt(af->fd, WAVE_FORMAT_PCM);  /* Size */
   outshrt(af->fd, achans);        /* Number of channels */
   outlong(af->fd, rate);          /* SamplesPerSec */
   outlong(af->fd, sampsize*rate); /* AvgBytesPerSec */
   outshrt(af->fd, sampsize);      /* BlockAlign */
   outshrt(af->fd, abytes*8);      /* BitsPerSample */

   /* Finish stream list */

   AVI_finish_list(af, "strl");

   }

   /* Finish header list */

   AVI_finish_list(af, "hdrl");

   /* Calculate the needed amount of junk bytes, output junk */

   pos = tell(af->fd);
   njunk = DATASTART - pos - 8 - 12;
   out4cc (af->fd, "JUNK");
   outlong(af->fd, njunk);
   // for(i=0;i<njunk;i++) fputc(0,af->fd);

}
void AVI_add_frame(AVI_file *af, char *jpeg_data, int length)
{
   int nw;

   if(af->n_idx>=MAXIDX) {fprintf(stderr,"IDX"); exit(0);}
   af->idx[af->n_idx  ] = MAKEFOURCC('0','0','d','b');
   af->idx[af->n_idx+1] = 0x10; /* RJ: No idea what that means */
   af->idx[af->n_idx+2] = tell(af->fd);
   af->idx[af->n_idx+3] = length;
   af->n_idx += 4;

   out4cc(af->fd, "00db");
   outlong(af->fd, length);

   nw = write(af->fd,jpeg_data,length);
   if(nw != length) { perror("write"); exit(1); }

   af->num_of_frames++;
}
void AVI_add_audio(AVI_file *af, char *audio_data, int length)
{
   int nw;

   if(af->n_idx>=MAXIDX) {fprintf(stderr,"IDX"); exit(0);}
   af->idx[af->n_idx  ] = MAKEFOURCC('0','1','w','b');
   af->idx[af->n_idx+1] = 0x00; /* RJ: No idea what that means */
   af->idx[af->n_idx+2] = tell(af->fd);
   af->idx[af->n_idx+3] = length;
   af->n_idx += 4;

   out4cc(af->fd, "01wb");
   outlong(af->fd, length);

   nw = write(af->fd,audio_data,length);
   if(nw != length) { perror("write"); exit(1); }

   af->num_of_audio_bytes += length;
}

/*******************************************************************
 *                                                                 *
 *    Utilities for reading video and audio from an AVI File       *
 *                                                                 *
 *******************************************************************/

/* Some FOURCCs */

#define RIFFtag MAKEFOURCC('R','I','F','F')
#define LISTtag MAKEFOURCC('L','I','S','T')
#define strhtag MAKEFOURCC('s','t','r','h')
#define strftag MAKEFOURCC('s','t','r','f')

/* The error codes delivered by avi_read_header */

#define ERR_EOF     1
#define ERR_NO_AVI  2
#define ERR_NO_MJPG 3
#define ERR_NO_PCM  4
#define ERR_NO_VIDS 5

/* The following variables are exported for interpretation after avi_read_header */

double AVI_fps     = 0;
long   AVI_width   = 0;
long   AVI_height  = 0;
long   AVI_a_chans = 0;
long   AVI_a_rate  = 0;
long   AVI_a_bits  = 0;


int avi_read_header(FILE * avifile)
{

   int tag, n, name;
   int lasttag = 0;
   int vids_strh_seen = 0;
   int vids_strf_seen = 0;
   unsigned long args[64];

   AVI_a_chans = 0; /* No audio */

   /* Read first 12 bytes and check that this is an AVI file */

   if ( fread(&tag ,4,1,avifile) != 1 ||
        fread(&n   ,4,1,avifile) != 1 ||
        fread(&name,4,1,avifile) != 1 )  return ERR_EOF;

   if(tag != RIFFtag || name != MAKEFOURCC('A','V','I',' ')) return ERR_NO_AVI;


   /* Read all tags until we encounter a 'LIST xxx movi' tag */

   while(1)
   {
      if( fread(&tag,4,1,avifile) != 1 ||
          fread(&n  ,4,1,avifile) != 1 ) return ERR_EOF;
      if(n&1) n++; /* Odd values are padded */

      /* Unless a LIST tag hasn't name movi, it is ignored */

      if ( tag == LISTtag )
      {
         if( fread(&name,4,1,avifile)!= 1 ) return ERR_EOF;
         if(name==MAKEFOURCC('m','o','v','i')) return 0;
         continue;
      }

      /* Read the arguments of this tag, 256 bytes are sufficient,
         remainder is skipped */

      if(n<=256)
      {
         if(fread(args,  n,1,avifile) != 1 ) return ERR_EOF;
      }
      else
      {
         if(fread(args,256,1,avifile) != 1 ) return ERR_EOF;
         if(fseek(avifile,n-256,SEEK_CUR) )  return ERR_EOF;
      }

      /* Interpret the tag and its args */
      switch (tag)
      {
         case strhtag:
            if(args[0] == MAKEFOURCC('v','i','d','s'))
            {
               if(args[1] != MAKEFOURCC('M','J','P','G') &&
                  args[1] != MAKEFOURCC('m','j','p','g') ) return ERR_NO_MJPG;
               if(args[5]!=0) AVI_fps = (double)args[6] / (double)args[5];
               vids_strh_seen = 1;
               lasttag = 1; /* vids */
            }
            else if (args[0] == MAKEFOURCC('a','u','d','s'))
            {
               lasttag = 2; /* auds */
            }
            break;

         case strftag:
            if(lasttag == 1)
            {
               AVI_width  = args[1];
               AVI_height = args[2];
               if(args[4] != MAKEFOURCC('M','J','P','G') &&
                  args[4] != MAKEFOURCC('m','j','p','g') ) return ERR_NO_MJPG;
               vids_strf_seen = 1;
            }
            else if(lasttag == 2)
            {
               /* Check audio format (must be PCM) */
               if( (args[0]&0xffff) != 1) return ERR_NO_PCM;
               AVI_a_chans = args[0]>>16;
               AVI_a_rate  = args[1];
               AVI_a_bits  = args[3]>>16;
            }
            break;

         default:
            lasttag = 0;
            break;
      }
   }

   if(!vids_strh_seen || !vids_strf_seen) return ERR_NO_VIDS;
   return 0;
}

int avi_read_data(FILE * avifile, char *vidbuf, long max_vidbuf,
                                  char *audbuf, long max_audbuf,
                                  long *len)
{

/*
 * Return codes:
 *
 *    1 = video data read
 *    2 = audio data read
 *    0 = reached EOF
 *   -1 = video buffer too small
 *   -2 = audio buffer too small
 */

   int tag, n;

   while(1)
   {
      /* Read tag and length */

      if( fread(&tag,4,1,avifile) != 1 ||
          fread(&n  ,4,1,avifile) != 1 ) return 0;
      if(n&1) n++; /* Odd values are padded */

      switch(tag)
      {
         case MAKEFOURCC('0','0','d','b'):
         case MAKEFOURCC('0','0','d','c'):
            if(n>max_vidbuf) return -1;
            *len = n;
            if(fread(vidbuf,n,1,avifile) != 1 ) return 0;
            return 1;
            break;

         case MAKEFOURCC('0','1','w','b'):
            if(n>max_audbuf) return -2;
            *len = n;
            if(fread(audbuf,n,1,avifile) != 1 ) return 0;
            return 2;
            break;

         default:
            if(fseek(avifile,n,SEEK_CUR) )  return 0;
            break;
      }
   }
}

/*******************************************************************
 *                                                                 *
 *    Utility for checking JPEG data read                          *
 *                                                                 *
 *******************************************************************/

/*
 *  Scan the jpeg headers and find SOF marker (for extracting
 *  width and height) and APP0 marker.
 *  Written loosely like rdjpgcom.c by Thomas G. Lane in the
 *  Independent JPEG Group's software.
 *
 *  Copyright (C) 1999 Rainer Johanni <Rainer@Johanni.de>
 */

#define M_SOI   0xD8		/* Start Of Image (beginning of datastream) */
#define M_EOI   0xD9		/* End Of Image (end of datastream) */
#define M_SOS   0xDA		/* Start Of Scan (begins compressed data) */
#define M_COM   0xFE		/* COMment */

/*
 * Parse the marker stream until SOS or EOI is seen;
 */

int check_JPEG_header (unsigned char * jpgdata, int jpglen,
                       unsigned int * width, unsigned int * height,
                       unsigned char * app0)
{
   int marker, length;
   int data_precision, num_components;
   int c1, c2, i, p;

   /* p is the pointer within the jpeg data */

   p = 0;

   /* Zero app0 - has to have a length of 16 */

   for(i=0;i<16;i++) app0[i] = 0;

   /* Zero width and height - don't return garbage */

   *width  = 0;
   *height = 0;

   /* Read the initial marker, which should be SOI */

   c1 = jpgdata[p++];
   c2 = jpgdata[p++];
   if (c1 != 0xFF || c2 != M_SOI) return -1;

   /* Scan miscellaneous markers until we reach SOS */

   while(1)
   {
      /* get next marker */

      /* Find 0xFF byte; skip any non-FFs */
      while(jpgdata[p] != 0xFF)
      {
         p++;
         if(p>=jpglen) return -1;
      }

      /* Get marker code byte, swallowing any duplicate FF bytes */
      while(jpgdata[p] == 0xFF)
      {
         p++;
         if(p>=jpglen) return -1;
      }

      marker = jpgdata[p++];

      /* Markers 0xc0 ... 0xc2 are SOF markers, extract dimensions */
      /* There are more SOF markers for lossless JPEG etc., we don't
         care about those */

      if( marker >= 0xc0 && marker <= 0xc2 )
      {
         if (p>=jpglen-8) return -1;
         c1 = jpgdata[p++];
         c2 = jpgdata[p++];
         length = (c1<<8) + c2; /* usual parameter length count */

         data_precision = jpgdata[p++];
         c1 = jpgdata[p++];
         c2 = jpgdata[p++];
         *height = (c1<<8) + c2;
         c1 = jpgdata[p++];
         c2 = jpgdata[p++];
         *width  = (c1<<8) + c2;
         num_components = jpgdata[p++];

         if (length != (unsigned int) (8 + num_components * 3)) return -1;

         if (p>=jpglen-num_components*3) return -1;

         p += num_components*3;

         continue;
      }

      if(marker == M_SOS) return 0; /* stop before hitting compressed data */

      if(marker == M_EOI) return 0; /* in case it's a tables-only JPEG stream */

      /* APP0 marker */

      if(marker == 0xe0 )
      {
         if (p>=jpglen-2) return -1;
         c1 = jpgdata[p++];
         c2 = jpgdata[p++];
         length = (c1<<8) + c2; /* usual parameter length count */

         length -= 2; /* remaining data */

         if (p>=jpglen-length) return -1;

         for(i=0;i<length;i++)
         {
            if(i<16)
               app0[i] = jpgdata[p++];
            else
               p++;
         }
         continue;
      }

      /* If we reach this point, the marker was not identified */
      /* Skip it, assuming it has a parameter count... */

      if (p>=jpglen-2) return -1;
      c1 = jpgdata[p++];
      c2 = jpgdata[p++];
      length = (c1<<8) + c2; /* usual parameter length count */

      length -= 2; /* remaining data */

      if (p>=jpglen-length) return -1;

      p += length;

   } /* end while loop */
}
